# Code from https://github.com/ros-infrastructure/ros_buildfarm
# (c) Open Source Robotics Foundation
import ast
import logging
from jenkinsapi.utils.requester import Requester

logger = logging.getLogger(__name__)


class CrumbRequester(Requester):
    """Adapter for Requester inserting the crumb in every request."""

    def __init__(self, *args, **kwargs):
        super(CrumbRequester, self).__init__(*args, **kwargs)
        self._baseurl = kwargs["baseurl"]
        self._last_crumb_data = None

    def post_url(
        self,
        url,
        params=None,
        data=None,
        files=None,
        headers=None,
        allow_redirects=True,
        **kwargs,
    ):
        if self._last_crumb_data:
            # first try request with previous crumb if available
            response = self._post_url_with_crumb(
                self._last_crumb_data,
                url,
                params,
                data,
                files,
                headers,
                allow_redirects,
                **kwargs,
            )
            # code 403 might indicate that the crumb is not valid anymore
            if response.status_code != 403:
                return response

        # fetch new crumb (if server has crumbs enabled)
        if self._last_crumb_data is not False:
            self._last_crumb_data = self._get_crumb_data()

        return self._post_url_with_crumb(
            self._last_crumb_data,
            url,
            params,
            data,
            files,
            headers,
            allow_redirects,
            **kwargs,
        )

    def _get_crumb_data(self):
        response = self.get_url(self._baseurl + "/crumbIssuer/api/python")
        if response.status_code in [404]:
            logger.warning("The Jenkins master does not require a crumb")
            return False
        if response.status_code not in [200]:
            raise RuntimeError("Failed to fetch crumb: %s" % response.text)
        crumb_issuer_response = ast.literal_eval(response.text)
        crumb_request_field = crumb_issuer_response["crumbRequestField"]
        crumb = crumb_issuer_response["crumb"]
        logger.debug("Fetched crumb: %s", crumb)
        return {crumb_request_field: crumb}

    def _post_url_with_crumb(
        self,
        crumb_data,
        url,
        params,
        data,
        files,
        headers,
        allow_redirects,
        **kwargs,
    ):
        if crumb_data:
            if headers is None:
                headers = crumb_data
            else:
                headers.update(crumb_data)

        return super(CrumbRequester, self).post_url(
            url, params, data, files, headers, allow_redirects, **kwargs
        )



"""
Module for jenkinsapi labels
"""

from jenkinsapi.jenkinsbase import JenkinsBase
import logging

log = logging.getLogger(__name__)


class Label(JenkinsBase):
    """
    Class to hold information on labels that tied to a collection of jobs
    """

    def __init__(self, baseurl, labelname, jenkins_obj):
        """
        Init a label object by providing all relevant pointers to it
        :param baseurl: basic url for querying information on a node
        :param labelname: name of the label
        :param jenkins_obj: ref to the jenkins obj
        :return: Label obj
        """
        self.labelname = labelname
        self.jenkins = jenkins_obj
        self.baseurl = baseurl
        JenkinsBase.__init__(self, baseurl)

    def __str__(self):
        return "%s" % (self.labelname)

    def get_jenkins_obj(self):
        return self.jenkins

    def is_online(self):
        return not self.poll(tree="offline")["offline"]

    def get_tied_jobs(self):
        """
        Get a list of jobs.
        """
        if self.get_tied_job_names():
            for job in self.get_tied_job_names():
                yield self.get_jenkins_obj().get_job(job["name"])

    def get_tied_job_names(self):
        """
        Get a list of the name of tied jobs.
        """
        return self.poll(tree="tiedJobs[name]")["tiedJobs"]



"""
Module for MutableJenkinsThing
"""


class MutableJenkinsThing(object):
    """
    A mixin for certain mutable objects which can be renamed and deleted.
    """

    def get_delete_url(self) -> str:
        return f"{self.baseurl}/doDelete"

    def get_rename_url(self) -> str:
        return f"{self.baseurl}/doRename"



"""
Module for converting jsonp to json.
"""

from __future__ import print_function


def jsonp_to_json(jsonp):
    try:
        l_index = jsonp.index("(") + 1
        r_index = jsonp.rindex(")")
    except ValueError:
        print("Input is not in jsonp format.")
        return None

    res = jsonp[l_index:r_index]
    return res



"""
Module for jenkinsapi Node class
"""

from __future__ import annotations

import json
import logging

import xml.etree.ElementTree as ET

import time
from jenkinsapi.jenkinsbase import JenkinsBase
from jenkinsapi.custom_exceptions import PostRequired, TimeOut
from jenkinsapi.custom_exceptions import JenkinsAPIException
from urllib.parse import quote as urlquote

log = logging.getLogger(__name__)


class Node(JenkinsBase):
    """
    Class to hold information on nodes that are attached as slaves
    to the master jenkins instance
    """

    def __init__(
        self,
        jenkins_obj: "Jenkins",
        baseurl: str,
        nodename: str,
        node_dict,
        poll: bool = True,
    ) -> None:
        """
        Init a node object by providing all relevant pointers to it
        :param jenkins_obj: ref to the jenkins obj
        :param baseurl: basic url for querying information on a node
            If url is not set - object will construct it itself. This is
            useful when node is being created and not exists in Jenkins yet
        :param nodename: hostname of the node
        :param dict node_dict: Dict with node parameters as described below
        :param bool poll: set to False if node does not exist or automatic
            refresh from Jenkins is not required. Default is True.
            If baseurl parameter is set to None - poll parameter will be
            set to False

        JNLP Node:
            {
                'num_executors': int,
                'node_description': str,
                'remote_fs': str,
                'labels': str,
                'exclusive': bool
            }

        SSH Node:
        {
            'num_executors': int,
            'node_description': str,
            'remote_fs': str,
            'labels': str,
            'exclusive': bool,
            'host': str,
            'port': int
            'credential_description': str,
            'jvm_options': str,
            'java_path': str,
            'prefix_start_slave_cmd': str,
            'suffix_start_slave_cmd': str
            'max_num_retries': int,
            'retry_wait_time': int,
            'retention': str ('Always' or 'OnDemand')
            'ondemand_delay': int (only for OnDemand retention)
            'ondemand_idle_delay': int (only for OnDemand retention)
            'env': [
                {
                    'key':'TEST',
                    'value':'VALUE'
                },
                {
                    'key':'TEST2',
                    'value':'value2'
                }
            ],
            'tool_location': [
                {
                    "key": "hudson.tasks.Maven$MavenInstallation$DescriptorImpl@Maven 3.0.5",  # noqa
                    "home": "/home/apache-maven-3.0.5/"
                },
                {
                    "key": "hudson.plugins.git.GitTool$DescriptorImpl@Default",
                    "home": "/home/git-3.0.5/"
                },
            ]
        }

        :return: None
        :return: Node obj
        """
        self.name: str = nodename
        self.jenkins: "Jenkins" = jenkins_obj
        if not baseurl:
            poll = False
            baseurl = f"{self.jenkins.baseurl}/computer/{self.name}"
        JenkinsBase.__init__(self, baseurl, poll=poll)
        self.node_attributes: dict = node_dict
        self._element_tree = None
        self._config = None

    def get_node_attributes(self) -> dict:
        """
        Gets node attributes as dict

        Used by Nodes object when node is created

        :return: Node attributes dict formatted for Jenkins API request
            to create node
        """
        na: dict = self.node_attributes
        if not na.get("credential_description", False):
            # If credentials description is not present - we will create
            # JNLP node
            launcher = {"stapler-class": "hudson.slaves.JNLPLauncher"}
        else:
            try:
                credential = self.jenkins.credentials[
                    na["credential_description"]
                ]
            except KeyError:
                raise JenkinsAPIException(
                    'Credential with description "%s"'
                    " not found" % na["credential_description"]
                )

            retries: int = (
                na["max_num_retries"] if "max_num_retries" in na else 0
            )
            re_wait: int = (
                na["retry_wait_time"] if "retry_wait_time" in na else 0
            )
            launcher = {
                "stapler-class": "hudson.plugins.sshslaves.SSHLauncher",
                "$class": "hudson.plugins.sshslaves.SSHLauncher",
                "host": na["host"],
                "port": na["port"],
                "credentialsId": credential.credential_id,
                "jvmOptions": na["jvm_options"],
                "javaPath": na["java_path"],
                "prefixStartSlaveCmd": na["prefix_start_slave_cmd"],
                "suffixStartSlaveCmd": na["suffix_start_slave_cmd"],
                "maxNumRetries": retries,
                "retryWaitTime": re_wait,
            }

        retention = {
            "stapler-class": "hudson.slaves.RetentionStrategy$Always",
            "$class": "hudson.slaves.RetentionStrategy$Always",
        }
        if "retention" in na and na["retention"].lower() == "ondemand":
            retention = {
                "stapler-class": "hudson.slaves.RetentionStrategy$Demand",
                "$class": "hudson.slaves.RetentionStrategy$Demand",
                "inDemandDelay": na["ondemand_delay"],
                "idleDelay": na["ondemand_idle_delay"],
            }

        node_props: dict = {"stapler-class-bag": "true"}
        if "env" in na:
            node_props.update(
                {
                    "hudson-slaves-EnvironmentVariablesNodeProperty": {
                        "env": na["env"]
                    }
                }
            )
        if "tool_location" in na:
            node_props.update(
                {
                    "hudson-tools-ToolLocationNodeProperty": {
                        "locations": na["tool_location"]
                    }
                }
            )

        params = {
            "name": self.name,
            "type": "hudson.slaves.DumbSlave$DescriptorImpl",
            "json": json.dumps(
                {
                    "name": self.name,
                    "nodeDescription": na.get("node_description", ""),
                    "numExecutors": na["num_executors"],
                    "remoteFS": na["remote_fs"],
                    "labelString": na["labels"],
                    "mode": "EXCLUSIVE" if na["exclusive"] else "NORMAL",
                    "retentionStrategy": retention,
                    "type": "hudson.slaves.DumbSlave",
                    "nodeProperties": node_props,
                    "launcher": launcher,
                }
            ),
        }

        return params

    def get_jenkins_obj(self) -> "Jenkins":
        return self.jenkins

    def __str__(self) -> str:
        return self.name

    def is_online(self) -> bool:
        return not self.poll(tree="offline")["offline"]

    def is_temporarily_offline(self) -> bool:
        return self.poll(tree="temporarilyOffline")["temporarilyOffline"]

    def is_jnlpagent(self) -> bool:
        return self._data["jnlpAgent"]

    def is_idle(self) -> bool:
        return self.poll(tree="idle")["idle"]

    def set_online(self) -> None:
        """
        Set node online.
        Before change state verify client state: if node set 'offline'
        but 'temporarilyOffline' is not set - client has connection problems
        and AssertionError raised.
        If after run node state has not been changed raise AssertionError.
        """
        self.poll()
        # Before change state check if client is connected
        if self._data["offline"] and not self._data["temporarilyOffline"]:
            raise AssertionError(
                "Node is offline and not marked as "
                "temporarilyOffline, check client "
                "connection: offline = %s, "
                "temporarilyOffline = %s"
                % (self._data["offline"], self._data["temporarilyOffline"])
            )

        if self._data["offline"] and self._data["temporarilyOffline"]:
            self.toggle_temporarily_offline()
            if self._data["offline"]:
                raise AssertionError(
                    "The node state is still offline, "
                    "check client connection:"
                    " offline = %s, "
                    "temporarilyOffline = %s"
                    % (self._data["offline"], self._data["temporarilyOffline"])
                )

    def set_offline(self, message="requested from jenkinsapi") -> None:
        """
        Set node offline.
        If after run node state has not been changed raise AssertionError.
        : param message: optional string explain why you are taking this
            node offline
        """
        if not self._data["offline"]:
            self.toggle_temporarily_offline(message)
            data = self.poll(tree="offline,temporarilyOffline")
            if not data["offline"]:
                raise AssertionError(
                    "The node state is still online:"
                    + "offline = %s , temporarilyOffline = %s"
                    % (data["offline"], data["temporarilyOffline"])
                )

    def launch(self) -> None:
        """
        Tries to launch a connection with the slave if it is currently
        disconnected. Because launching a connection with the slave does not
        mean it is online (a slave can be launched, but set offline), this
        function does not check if the launch was successful.
        """
        if not self._data["launchSupported"]:
            raise AssertionError("The node does not support manually launch.")

        if not self._data["manualLaunchAllowed"]:
            raise AssertionError(
                "It is not allowed to manually launch this node."
            )

        url = self.baseurl + "/launchSlaveAgent"
        html_result = self.jenkins.requester.post_and_confirm_status(
            url, data={}
        )
        log.debug(html_result)

    def toggle_temporarily_offline(
        self, message="requested from jenkinsapi"
    ) -> None:
        """
        Switches state of connected node (online/offline) and
        set 'temporarilyOffline' property (True/False)
        Calling the same method again will bring node status back.
        :param message: optional string can be used to explain why you
            are taking this node offline
        """
        initial_state = self.is_temporarily_offline()
        url = (
            self.baseurl + "/toggleOffline?offlineMessage=" + urlquote(message)
        )
        try:
            html_result = self.jenkins.requester.get_and_confirm_status(url)
        except PostRequired:
            html_result = self.jenkins.requester.post_and_confirm_status(
                url, data={}
            )

        self.poll()
        log.debug(html_result)
        state = self.is_temporarily_offline()
        if initial_state == state:
            raise AssertionError(
                "The node state has not changed: temporarilyOffline = %s"
                % state
            )

    def update_offline_reason(self, reason: str) -> None:
        """
        Update offline reason on a temporary offline clsuter
        """

        if self.is_temporarily_offline():
            url = (
                self.baseurl
                + "/changeOfflineCause?offlineMessage="
                + urlquote(reason)
            )
            self.jenkins.requester.post_and_confirm_status(url, data={})

    def offline_reason(self) -> str:
        return self._data["offlineCauseReason"]

    @property
    def _et(self):
        return self._get_config_element_tree()

    def _get_config_element_tree(self) -> ET.Element:
        """
        Returns an xml element tree for the node's config.xml. The
        resulting tree is cached for quick lookup.
        """
        if self._config is None:
            self.load_config()

        if self._element_tree is None:
            self._element_tree = ET.fromstring(self._config)
        return self._element_tree

    def get_config(self) -> str:
        """
        Returns the config.xml from the node.
        """
        response = self.jenkins.requester.get_and_confirm_status(
            "%(baseurl)s/config.xml" % self.__dict__
        )
        return response.text

    def load_config(self) -> None:
        """
        Loads the config.xml for the node allowing it to be re-queried
        without generating new requests.
        """
        if self.name == "Built-In Node":
            raise JenkinsAPIException("Built-In node does not have config.xml")

        self._config = self.get_config()
        self._get_config_element_tree()

    def upload_config(self, config_xml: str) -> None:
        """
        Uploads config_xml to the config.xml for the node.
        """
        if self.name == "Built-In Node":
            raise JenkinsAPIException("Built-In node does not have config.xml")

        self.jenkins.requester.post_and_confirm_status(
            "%(baseurl)s/config.xml" % self.__dict__, data=config_xml
        )

    def get_labels(self) -> str | None:
        """
        Returns the labels for a slave as a string with each label
        separated by the ' ' character.
        """
        return self.get_config_element("label")

    def get_num_executors(self) -> str:
        try:
            return self.get_config_element("numExecutors")
        except JenkinsAPIException:
            return self._data["numExecutors"]

    def set_num_executors(self, value: int | str) -> None:
        """
        Sets number of executors for node

        Warning! Setting number of executors on master node will erase all
        other settings
        """
        set_value = value if isinstance(value, str) else str(value)

        if self.name == "Built-In Node":
            # master node doesn't have config.xml, so we're going to submit
            # form here
            data = "json=%s" % urlquote(
                json.dumps(
                    {
                        "numExecutors": set_value,
                        "nodeProperties": {"stapler-class-bag": "true"},
                    }
                )
            )

            url = self.baseurl + "/configSubmit"
            self.jenkins.requester.post_and_confirm_status(url, data=data)
        else:
            self.set_config_element("numExecutors", set_value)

        self.poll()

    def get_config_element(self, el_name: str) -> str:
        """
        Returns simple config element.

        Better not to be used to return "nodeProperties" or "launcher"
        """
        return self._et.find(el_name).text

    def set_config_element(self, el_name: str, value: str) -> None:
        """
        Sets simple config element
        """
        self._et.find(el_name).text = value
        xml_str = ET.tostring(self._et)
        self.upload_config(xml_str)

    def get_monitor(self, monitor_name: str, poll_monitor=True) -> str:
        """
        Polls the node returning one of the monitors in the monitorData
        branch of the returned node api tree.
        """
        monitor_data_key = "monitorData"
        if poll_monitor:
            # polling as monitors like response time can be updated
            monitor_data = self.poll(tree=monitor_data_key)[monitor_data_key]
        else:
            monitor_data = self._data[monitor_data_key]

        full_monitor_name = "hudson.node_monitors.{0}".format(monitor_name)
        if full_monitor_name not in monitor_data:
            raise AssertionError("Node monitor %s not found" % monitor_name)

        return monitor_data[full_monitor_name]

    def get_available_physical_memory(self) -> int:
        """
        Returns the node's available physical memory in bytes.
        """
        monitor_data = self.get_monitor("SwapSpaceMonitor")
        return monitor_data["availablePhysicalMemory"]

    def get_available_swap_space(self) -> int:
        """
        Returns the node's available swap space in bytes.
        """
        monitor_data = self.get_monitor("SwapSpaceMonitor")
        return monitor_data["availableSwapSpace"]

    def get_total_physical_memory(self) -> int:
        """
        Returns the node's total physical memory in bytes.
        """
        monitor_data = self.get_monitor("SwapSpaceMonitor")
        return monitor_data["totalPhysicalMemory"]

    def get_total_swap_space(self) -> int:
        """
        Returns the node's total swap space in bytes.
        """
        monitor_data = self.get_monitor("SwapSpaceMonitor")
        return monitor_data["totalSwapSpace"]

    def get_workspace_path(self) -> str:
        """
        Returns the local path to the node's Jenkins workspace directory.
        """
        monitor_data = self.get_monitor("DiskSpaceMonitor")
        return monitor_data["path"]

    def get_workspace_size(self) -> int:
        """
        Returns the size in bytes of the node's Jenkins workspace directory.
        """
        monitor_data = self.get_monitor("DiskSpaceMonitor")
        return monitor_data["size"]

    def get_temp_path(self) -> str:
        """
        Returns the local path to the node's temp directory.
        """
        monitor_data = self.get_monitor("TemporarySpaceMonitor")
        return monitor_data["path"]

    def get_temp_size(self) -> int:
        """
        Returns the size in bytes of the node's temp directory.
        """
        monitor_data = self.get_monitor("TemporarySpaceMonitor")
        return monitor_data["size"]

    def get_architecture(self) -> str:
        """
        Returns the system architecture of the node eg. "Linux (amd64)".
        """
        # no need to poll as the architecture will never change
        return self.get_monitor("ArchitectureMonitor", poll_monitor=False)

    def block_until_idle(self, timeout: int, poll_time: int = 5) -> None:
        """
        Blocks until the node become idle.
        :param timeout: Time in second when the wait is aborted.
        :param poll_time: Interval in seconds between each check.
        :@raise TimeOut
        """
        start_time = time.time()
        while not self.is_idle() and (time.time() - start_time) < timeout:
            log.debug(
                "Waiting for the node to become idle. Elapsed time: %s",
                (time.time() - start_time),
            )
            time.sleep(poll_time)

        if not self.is_idle():
            raise TimeOut(
                "The node has not become idle after {} minutes.".format(
                    timeout / 60
                )
            )

    def get_response_time(self) -> int:
        """
        Returns the node's average response time.
        """
        monitor_data = self.get_monitor("ResponseTimeMonitor")
        return monitor_data["average"]

    def get_clock_difference(self) -> int:
        """
        Returns the difference between the node's clock and
        the master Jenkins clock.
        Used to detect out of sync clocks.
        """
        monitor_data = self.get_monitor("ClockMonitor")
        return monitor_data["diff"]



"""
Module for jenkinsapi nodes
"""

from __future__ import annotations

from typing import Iterator

import logging

from urllib.parse import urlencode
from jenkinsapi.node import Node
from jenkinsapi.jenkinsbase import JenkinsBase
from jenkinsapi.custom_exceptions import JenkinsAPIException
from jenkinsapi.custom_exceptions import UnknownNode
from jenkinsapi.custom_exceptions import PostRequired

log: logging.Logger = logging.getLogger(__name__)


class Nodes(JenkinsBase):
    """
    Class to hold information on a collection of nodes
    """

    def __init__(self, baseurl: str, jenkins_obj: "Jenkins") -> None:
        """
        Handy access to all of the nodes on your Jenkins server
        """
        self.jenkins = jenkins_obj
        JenkinsBase.__init__(
            self,
            (
                baseurl.rstrip("/")
                if "/computer" in baseurl
                else baseurl.rstrip("/") + "/computer"
            ),
        )

    def get_jenkins_obj(self) -> "Jenkins":
        return self.jenkins

    def __str__(self) -> str:
        return "Nodes @ %s" % self.baseurl

    def __contains__(self, node_name: str) -> bool:
        return node_name in self.keys()

    def iterkeys(self) -> Iterator[str]:
        """
        Return an iterator over the container's node names.

        Using iterkeys() while creating nodes may raise a RuntimeError
        or fail to iterate over all entries.
        """
        for item in self._data["computer"]:
            yield item["displayName"]

    def keys(self) -> list[str]:
        """
        Return a copy of the container's list of node names.
        """
        return list(self.iterkeys())

    def _make_node(self, nodename) -> Node:
        """
        Creates an instance of Node for the given nodename.
        This function assumes the returned node exists.
        """
        if nodename.lower() == "built-in node":
            nodeurl = "%s/(%s)" % (self.baseurl, "built-in")
        else:
            nodeurl = "%s/%s" % (self.baseurl, nodename)
        return Node(self.jenkins, nodeurl, nodename, node_dict={})

    def iteritems(self) -> Iterator[tuple[str, Node]]:
        """
        Return an iterator over the container's (name, node) pairs.

        Using iteritems() while creating nodes may raise a RuntimeError or
        fail to iterate over all entries.
        """
        for item in self._data["computer"]:
            nodename = item["displayName"]
            try:
                yield nodename, self._make_node(nodename)
            except Exception:
                raise JenkinsAPIException("Unable to iterate nodes")

    def items(self) -> list[tuple[str, Node]]:
        """
        Return a copy of the container's list of (name, node) pairs.
        """
        return list(self.iteritems())

    def itervalues(self) -> Iterator[Node]:
        """
        Return an iterator over the container's nodes.

        Using itervalues() while creating nodes may raise a RuntimeError
        or fail to iterate over all entries.
        """
        for item in self._data["computer"]:
            try:
                yield self._make_node(item["displayName"])
            except Exception:
                raise JenkinsAPIException("Unable to iterate nodes")

    def values(self) -> list[Node]:
        """
        Return a copy of the container's list of nodes.
        """
        return list(self.itervalues())

    def __getitem__(self, nodename: str) -> Node:
        if nodename in self:
            return self._make_node(nodename)
        raise UnknownNode(nodename)

    def __len__(self) -> int:
        return len(self.keys())

    def __delitem__(self, item: str) -> None:
        if item in self and item != "Built-In Node":
            url = "%s/doDelete" % self[item].baseurl
            try:
                self.jenkins.requester.get_and_confirm_status(url)
            except PostRequired:
                # Latest Jenkins requires POST here. GET kept for compatibility
                self.jenkins.requester.post_and_confirm_status(url, data={})
            self.poll()
        else:
            if item != "Built-In Node":
                raise UnknownNode("Node %s does not exist" % item)

            log.info("Requests to remove built-in node ignored")

    def __setitem__(self, name: str, node_dict: dict):
        if not isinstance(node_dict, dict):
            raise ValueError('"node_dict" parameter must be a Node dict')
        if name not in self:
            self.create_node(name, node_dict)
        self.poll()

    def create_node(self, name: str, node_dict: dict) -> Node:
        """
        Create a new slave node

        :param str name: name of slave
        :param dict node_dict: node dict (See Node class)
        :return: node obj
        """
        if name in self:
            return self[name]

        node = Node(
            jenkins_obj=self.jenkins,
            baseurl="",
            nodename=name,
            node_dict=node_dict,
            poll=False,
        )

        url = "%s/computer/doCreateItem?%s" % (
            self.jenkins.baseurl,
            urlencode(node.get_node_attributes()),
        )
        data = {"json": urlencode(node.get_node_attributes())}
        self.jenkins.requester.post_and_confirm_status(url, data=data)
        self.poll()
        return self[name]

    def create_node_with_config(self, name: str, config: dict) -> Node | None:
        """
        Create a new slave node with specific configuration.
        Config should be resemble the output of node.get_node_attributes()
        :param str name: name of slave
        :param dict config: Node attributes for Jenkins API request
            to create node
            (See function output Node.get_node_attributes())
        :return: node obj
        """
        if name in self:
            return self[name]

        if not isinstance(config, dict):
            return None
        url = "%s/computer/doCreateItem?%s" % (
            self.jenkins.baseurl,
            urlencode(config),
        )
        data = {"json": urlencode(config)}
        self.jenkins.requester.post_and_confirm_status(url, data=data)
        self.poll()
        return self[name]



#!/usr/bin/env python
from __future__ import print_function

try:
    from SimpleHTTPServer import SimpleHTTPRequestHandler
except ImportError:
    from http.server import SimpleHTTPRequestHandler

try:
    import SocketServer as socketserver
except ImportError:
    import socketserver

import logging
import cgi

PORT = 8080


class ServerHandler(SimpleHTTPRequestHandler):
    def do_GET(self):
        logging.error(self.headers)
        super().do_GET()

    def do_POST(self):
        logging.error(self.headers)
        form = cgi.FieldStorage(
            fp=self.rfile,
            headers=self.headers,
            environ={
                "REQUEST_METHOD": "POST",
                "CONTENT_TYPE": self.headers["Content-Type"],
            },
        )
        for item in form.list:
            logging.error(item)
        super().do_GET()


if __name__ == "__main__":
    Handler = ServerHandler

    httpd = socketserver.TCPServer(("", PORT), Handler)

    print("serving at port", PORT)
    httpd.serve_forever()



"""Module for custom_exceptions.

Where possible we try to throw exceptions with non-generic,
meaningful names.
"""


class JenkinsAPIException(Exception):
    """Base class for all errors"""

    pass


class NotFound(JenkinsAPIException):
    """Resource cannot be found"""

    pass


class ArtifactsMissing(NotFound):
    """Cannot find a build with all of the required artifacts."""

    pass


class UnknownJob(KeyError, NotFound):
    """Jenkins does not recognize the job requested."""

    pass


class UnknownView(KeyError, NotFound):
    """Jenkins does not recognize the view requested."""

    pass


class UnknownNode(KeyError, NotFound):
    """Jenkins does not recognize the node requested."""

    pass


class UnknownQueueItem(KeyError, NotFound):
    """Jenkins does not recognize the requested queue item"""

    pass


class UnknownPlugin(KeyError, NotFound):
    """Jenkins does not recognize the plugin requested."""

    pass


class NoBuildData(NotFound):
    """A job has no build data."""

    pass


class NotBuiltYet(NotFound):
    """A job has no build data."""

    pass


class ArtifactBroken(JenkinsAPIException):
    """An artifact is broken, wrong"""

    pass


class TimeOut(JenkinsAPIException):
    """Some jobs have taken too long to complete."""

    pass


class NoResults(JenkinsAPIException):
    """A build did not publish any results."""

    pass


class FailedNoResults(NoResults):
    """A build did not publish any results because it failed"""

    pass


class BadURL(ValueError, JenkinsAPIException):
    """A URL appears to be broken"""

    pass


class NotAuthorized(JenkinsAPIException):
    """Not Authorized to access resource"""

    # Usually thrown when we get a 403 returned
    pass


class NotSupportSCM(JenkinsAPIException):
    """
    It's a SCM that does not supported by current version of jenkinsapi
    """

    pass


class NotConfiguredSCM(JenkinsAPIException):
    """It's a job that doesn't have configured SCM"""

    pass


class NotInQueue(JenkinsAPIException):
    """It's a job that is not in the queue"""

    pass


class PostRequired(JenkinsAPIException):
    """Method requires POST and not GET"""

    pass


class BadParams(JenkinsAPIException):
    """Invocation was given bad or inappropriate params"""

    pass


class AlreadyExists(JenkinsAPIException):
    """
    Method requires POST and not GET
    """

    pass



"""
This module implements the Credentials class, which is intended to be a
container-like interface for all of the Global credentials defined on a single
Jenkins node.
"""

from __future__ import annotations

from typing import Iterator

import logging
from urllib.parse import urlencode

from jenkinsapi.credential import Credential
from jenkinsapi.credential import UsernamePasswordCredential
from jenkinsapi.credential import SecretTextCredential
from jenkinsapi.credential import SSHKeyCredential
from jenkinsapi.jenkinsbase import JenkinsBase
from jenkinsapi.custom_exceptions import JenkinsAPIException

log: logging.Logger = logging.getLogger(__name__)


class Credentials(JenkinsBase):
    """
    This class provides a container-like API which gives
    access to all global credentials on a Jenkins node.

    Returns a list of Credential Objects.
    """

    def __init__(self, baseurl: str, jenkins_obj: "Jenkins"):
        self.baseurl: str = baseurl
        self.jenkins: "Jenkins" = jenkins_obj
        JenkinsBase.__init__(self, baseurl)

        self.credentials = self._data["credentials"]

    def _poll(self, tree=None):
        url: str = self.python_api_url(self.baseurl) + "?depth=2"
        data = self.get_data(url, tree=tree)
        credentials = data["credentials"]
        for cred_id, cred_dict in credentials.items():
            cred_dict["credential_id"] = cred_id
            credentials[cred_id] = self._make_credential(cred_dict)

        return data

    def __str__(self) -> str:
        return "Global Credentials @ %s" % self.baseurl

    def get_jenkins_obj(self) -> "Jenkins":
        return self.jenkins

    def __iter__(self) -> Iterator[Credential]:
        for cred in self.credentials.values():
            yield cred.description

    def __contains__(self, description: str) -> bool:
        return description in self.keys()

    def iterkeys(self):
        return self.__iter__()

    def keys(self):
        return list(self.iterkeys())

    def iteritems(self) -> Iterator[str, "Credential"]:
        for cred in self.credentials.values():
            yield cred.description, cred

    def __getitem__(self, description: str) -> "Credential":
        for cred in self.credentials.values():
            if cred.description == description:
                return cred

        raise KeyError(
            'Credential with description "%s" not found' % description
        )

    def __len__(self) -> int:
        return len(self.keys())

    def __setitem__(self, description: str, credential: "Credential"):
        """
        Creates Credential in Jenkins using username, password and description
        Description must be unique in Jenkins instance
        because it is used to find Credential later.

        If description already exists - this method is going to update
        existing Credential

        :param str description: Credential description
        :param tuple credential_tuple: (username, password, description) tuple.
        """
        if description not in self:
            params = credential.get_attributes()
            url = "%s/createCredentials" % self.baseurl
            try:
                self.jenkins.requester.post_and_confirm_status(
                    url, params={}, data=urlencode(params)
                )
            except JenkinsAPIException as jae:
                raise JenkinsAPIException(
                    "Latest version of Credentials "
                    "plugin is required to be able "
                    "to create credentials. "
                    "Original exception: %s" % str(jae)
                )
        else:
            cred_id = self[description].credential_id
            credential.credential_id = cred_id
            params = credential.get_attributes_xml()
            url = "%s/credential/%s/config.xml" % (self.baseurl, cred_id)
            try:
                self.jenkins.requester.post_xml_and_confirm_status(
                    url, params={}, data=params
                )
            except JenkinsAPIException as jae:
                raise JenkinsAPIException(
                    "Latest version of Credentials "
                    "plugin is required to be able "
                    "to update credentials. "
                    "Original exception: %s" % str(jae)
                )

        self.poll()
        self.credentials = self._data["credentials"]
        if description not in self:
            raise JenkinsAPIException("Problem creating/updating credential.")

    def get(self, item, default):
        return self[item] if item in self else default

    def __delitem__(self, description: str):
        if description not in self:
            raise KeyError(
                'Credential with description "%s" not found' % description
            )
        params = {"Submit": "OK", "json": {}}
        url = "%s/credential/%s/doDelete" % (
            self.baseurl,
            self[description].credential_id,
        )
        try:
            self.jenkins.requester.post_and_confirm_status(
                url, params={}, data=urlencode(params)
            )
        except JenkinsAPIException as jae:
            raise JenkinsAPIException(
                "Latest version of Credentials "
                "required to be able to create "
                "credentials. Original exception: %s" % str(jae)
            )
        self.poll()
        self.credentials = self._data["credentials"]
        if description in self:
            raise JenkinsAPIException("Problem deleting credential.")

    def _make_credential(self, cred_dict):
        if cred_dict["typeName"] == "Username with password":
            cr = UsernamePasswordCredential(cred_dict)
        elif cred_dict["typeName"] == "SSH Username with private key":
            cr = SSHKeyCredential(cred_dict)
        elif cred_dict["typeName"] == "Secret text":
            cr = SecretTextCredential(cred_dict)
        else:
            cr = Credential(cred_dict)

        return cr


class Credentials2x(Credentials):
    """
    This class provides a container-like API which gives
    access to all global credentials on a Jenkins node.

    Returns a list of Credential Objects.
    """

    def _poll(self, tree=None):
        url = self.python_api_url(self.baseurl) + "?depth=2"
        data = self.get_data(url, tree=tree)
        credentials = data["credentials"]
        new_creds = {}
        for cred_dict in credentials:
            cred_dict["credential_id"] = cred_dict["id"]
            new_creds[cred_dict["id"]] = self._make_credential(cred_dict)

        data["credentials"] = new_creds

        return data


class CredentialsById(Credentials2x):
    """
    This class provides a container-like API which gives
    access to all global credentials on a Jenkins node.

    Returns a list of Credential Objects.
    """

    def __iter__(self):
        for cred in self.credentials.values():
            yield cred.credential_id

    def __contains__(self, credential_id):
        return credential_id in self.keys()

    def iteritems(self):
        for cred in self.credentials.values():
            yield cred.credential_id, cred

    def __getitem__(self, credential_id):
        for cred in self.credentials.values():
            if cred.credential_id == credential_id:
                return cred

        raise KeyError(
            'Credential with credential_id "%s" not found' % credential_id
        )



"""
This module implements the Jobs class, which is intended to be a container-like
interface for all of the jobs defined on a single Jenkins server.
"""

from __future__ import annotations

from typing import Iterator
import logging
import time

from jenkinsapi.job import Job
from jenkinsapi.custom_exceptions import JenkinsAPIException, UnknownJob

log = logging.getLogger(__name__)


class Jobs(object):
    """
    This class provides a container-like API which gives
    access to all jobs defined on the Jenkins server. It behaves
    like a dict in which keys are Job-names and values are actual
    jenkinsapi.Job objects.
    """

    def __init__(self, jenkins: "Jenkins") -> None:
        self.jenkins = jenkins
        self._data = []

    def _del_data(self, job_name: str) -> None:
        if not self._data:
            return
        for num, job_data in enumerate(self._data):
            if job_data["name"] == job_name:
                del self._data[num]
                return

    def __len__(self) -> int:
        return len(self.keys())

    def poll(self, tree="jobs[name,color,url]"):
        return self.jenkins.poll(tree=tree)

    def __delitem__(self, job_name: str) -> None:
        """
        Delete a job by name
        :param str job_name: name of a existing job
        :raises JenkinsAPIException:  When job is not deleted
        """
        if job_name in self:
            try:
                delete_job_url = self[job_name].get_delete_url()
                self.jenkins.requester.post_and_confirm_status(
                    delete_job_url, data="some random bytes..."
                )

                self._del_data(job_name)
            except JenkinsAPIException:
                # Sometimes jenkins throws NPE when removing job
                # It removes job ok, but it is good to be sure
                # so we re-try if job was not deleted
                if job_name in self:
                    delete_job_url = self[job_name].get_delete_url()
                    self.jenkins.requester.post_and_confirm_status(
                        delete_job_url, data="some random bytes..."
                    )
                    self._del_data(job_name)

    def __setitem__(self, key: str, value: str) -> "Job":
        """
        Create Job

        :param str key:     Job name
        :param str value:   XML configuration of the job

        .. code-block:: python
        api = Jenkins('http://localhost:8080/')
        new_job = api.jobs['my_new_job'] = config_xml
        """
        return self.create(key, value)

    def __getitem__(self, job_name: str) -> "Job":
        if job_name in self:
            job_data = [
                job_row
                for job_row in self._data
                if job_row["name"] == job_name
                or Job.get_full_name_from_url_and_baseurl(
                    job_row["url"], self.jenkins.baseurl
                )
                == job_name
            ][0]
            return Job(job_data["url"], job_data["name"], self.jenkins)
        else:
            raise UnknownJob(job_name)

    def iteritems(self) -> Iterator[str, "Job"]:
        """
        Iterate over the names & objects for all jobs
        """
        for job in self.itervalues():
            if job.name != job.get_full_name():
                yield job.get_full_name(), job
            else:
                yield job.name, job

    def __contains__(self, job_name: str) -> bool:
        """
        True if job_name exists in Jenkins
        """
        return job_name in self.keys()

    def iterkeys(self) -> Iterator[str]:
        """
        Iterate over the names of all available jobs
        """
        if not self._data:
            self._data = self.poll().get("jobs", [])
        for row in self._data:
            if row["name"] != Job.get_full_name_from_url_and_baseurl(
                row["url"], self.jenkins.baseurl
            ):
                yield Job.get_full_name_from_url_and_baseurl(
                    row["url"], self.jenkins.baseurl
                )
            else:
                yield row["name"]

    def itervalues(self) -> Iterator["Job"]:
        """
        Iterate over all available jobs
        """
        if not self._data:
            self._data = self.poll().get("jobs", [])
        for row in self._data:
            yield Job(row["url"], row["name"], self.jenkins)

    def keys(self) -> list[str]:
        """
        Return a list of the names of all jobs
        """
        return list(self.iterkeys())

    def create(self, job_name: str, config: str | bytes) -> "Job":
        """
        Create a job

        :param str jobname: Name of new job
        :param str config: XML configuration of new job
        :returns Job: new Job object
        """
        if job_name in self:
            return self[job_name]

        if not config:
            raise JenkinsAPIException("Job XML config cannot be empty")

        params = {"name": job_name}
        if isinstance(config, bytes):
            config = config.decode("utf-8")

        self.jenkins.requester.post_xml_and_confirm_status(
            self.jenkins.get_create_url(), data=config, params=params
        )
        # Reset to get it refreshed from Jenkins
        self._data = []

        return self[job_name]

    def create_multibranch_pipeline(
        self, job_name: str, config: str, block: bool = True, delay: int = 60
    ) -> list["Job"]:
        """
        Create a multibranch pipeline job

        :param str jobname: Name of new job
        :param str config: XML configuration of new job
        :param block: block until scan is finished?
        :param delay: max delay to wait for scan to finish (seconds)
        :returns list of new Jobs after scan
        """
        if not config:
            raise JenkinsAPIException("Job XML config cannot be empty")

        params = {"name": job_name}
        if isinstance(config, bytes):
            config = config.decode("utf-8")

        self.jenkins.requester.post_xml_and_confirm_status(
            self.jenkins.get_create_url(), data=config, params=params
        )
        # Reset to get it refreshed from Jenkins
        self._data = []

        # Launch a first scan / indexing to discover the branches...
        self.jenkins.requester.post_and_confirm_status(
            "{}/job/{}/build".format(self.jenkins.baseurl, job_name),
            data="",
            valid=[200, 302],  # expect 302 without redirects
            allow_redirects=False,
        )

        start_time = time.time()
        # redirect-url does not work with indexing;
        # so the only workaround found is to parse the console output
        # until scan has finished.
        scan_finished = False
        while not scan_finished and block and time.time() < start_time + delay:
            indexing_console_text = self.jenkins.requester.get_url(
                "{}/job/{}/indexing/consoleText".format(
                    self.jenkins.baseurl, job_name
                )
            )
            if (
                indexing_console_text.text.strip()
                .split("\n")[-1]
                .startswith("Finished:")
            ):
                scan_finished = True
            time.sleep(1)

        # now search for all jobs created; those who start with job_name + '/'
        jobs = []
        for name in self.jenkins.get_jobs_list():
            if name.startswith(job_name + "/"):
                jobs.append(self[name])

        return jobs

    def copy(self, job_name: str, new_job_name: str) -> "Job":
        """
        Copy a job
        :param str job_name: Name of an existing job
        :param new_job_name: Name of new job
        :returns Job: new Job object
        """
        params = {"name": new_job_name, "mode": "copy", "from": job_name}

        self.jenkins.requester.post_and_confirm_status(
            self.jenkins.get_create_url(), params=params, data=""
        )

        self._data = []

        return self[new_job_name]

    def rename(self, job_name: str, new_job_name: str) -> "Job":
        """
        Rename a job

        :param str job_name: Name of an existing job
        :param str new_job_name: Name of new job
        :returns Job: new Job object
        """
        params = {"newName": new_job_name}
        rename_job_url = self[job_name].get_rename_url()
        self.jenkins.requester.post_and_confirm_status(
            rename_job_url, params=params, data=""
        )

        self._data = []

        return self[new_job_name]

    def build(self, job_name: str, params=None, **kwargs) -> "QueueItem":
        """
        Executes build of a job

        :param str job_name:    Job name
        :param dict params:     Job parameters
        :param kwargs:          Parameters for Job.invoke() function
        :returns QueueItem:     Object to track build progress
        """
        if params:
            assert isinstance(params, dict)
            return self[job_name].invoke(build_params=params, **kwargs)

        return self[job_name].invoke(**kwargs)



"""
This module enables Manifest file parsing.
Copied from
https://chromium.googlesource.com/external/googleappengine/python/+/master
/google/appengine/tools/jarfile.py
"""

import zipfile

_MANIFEST_NAME = "META-INF/MANIFEST.MF"


class InvalidJarError(Exception):
    """
    InvalidJar exception class
    """

    pass


class Manifest(object):
    """The parsed manifest from a jar file.
    Attributes:
      main_section: a dict representing the main (first)
        section of the manifest.
        Each key is a string that is an attribute, such as
        'Manifest-Version', and the corresponding value is a string that
        is the value of the attribute, such as '1.0'.
      sections: a dict representing the other sections of the manifest.
        Each key is a string that is the value of the 'Name' attribute for
        the section, and the corresponding value is a dict like the
        main_section one, for the other attributes.
    """

    def __init__(self, main_section, sections):
        self.main_section = main_section
        self.sections = sections


def read_manifest(jar_file_name):
    """Read and parse the manifest out of the given jar.
    Args:
      jar_file_name: the name of the jar from which the manifest is to be read.
    Returns:
      A parsed Manifest object, or None if the jar has no manifest.
    Raises:
      IOError: if the jar does not exist or cannot be read.
    """
    with zipfile.ZipFile(jar_file_name) as jar:
        try:
            manifest_string = jar.read(_MANIFEST_NAME).decode("UTF-8")
        except KeyError:
            return None
        return _parse_manifest(manifest_string)


def _parse_manifest(manifest_string):
    """Parse a Manifest object out of the given string.
    Args:
      manifest_string: a str or unicode that is the manifest contents.
    Returns:
      A Manifest object parsed out of the string.
    Raises:
      InvalidJarError: if the manifest is not well-formed.
    """
    manifest_string = "\n".join(manifest_string.splitlines()).rstrip("\n")
    section_strings = manifest_string.split("\n\n")
    parsed_sections = [_parse_manifest_section(s) for s in section_strings]
    main_section = parsed_sections[0]
    sections = dict()
    try:
        for entry in parsed_sections[1:]:
            sections[entry["Name"]] = entry
    except KeyError:
        raise InvalidJarError(
            "Manifest entry has no Name attribute: %s" % entry
        )
    return Manifest(main_section, sections)


def _parse_manifest_section(section):
    """Parse a dict out of the given manifest section string.
    Args:
      section: a str or unicode that is the manifest section.
        It looks something like this (without the >):
        > Name: section-name
        > Some-Attribute: some value
        > Another-Attribute: another value
    Returns:
      A dict where the keys are the attributes (here, 'Name', 'Some-Attribute',
      'Another-Attribute'), and the values are the corresponding
      attribute values.
    Raises:
      InvalidJarError: if the manifest section is not well-formed.
    """
    section = section.replace("\n ", "")
    try:
        return dict(line.split(": ", 1) for line in section.split("\n"))
    except ValueError:
        raise InvalidJarError("Invalid manifest %r" % section)



"""
Module for jenkinsapi Views
"""

import logging
import json
from jenkinsapi.view import View
from jenkinsapi.custom_exceptions import JenkinsAPIException

log = logging.getLogger(__name__)


class Views(object):
    """
    An abstraction on a Jenkins object's views
    """

    LIST_VIEW = "hudson.model.ListView"
    NESTED_VIEW = "hudson.plugins.nested_view.NestedView"
    CATEGORIZED_VIEW = (
        "org.jenkinsci.plugins.categorizedview.CategorizedJobsView"
    )
    MY_VIEW = "hudson.model.MyView"
    DASHBOARD_VIEW = "hudson.plugins.view.dashboard.Dashboard"
    PIPELINE_VIEW = (
        "au.com.centrumsystems.hudson."
        "plugin.buildpipeline.BuildPipelineView"
    )

    def __init__(self, jenkins):
        self.jenkins = jenkins
        self._data = None

    def poll(self, tree=None):
        self._data = self.jenkins.poll(
            tree="views[name,url]" if tree is None else tree
        )

    def __len__(self):
        return len(self.keys())

    def __delitem__(self, view_name):
        if view_name == "All":
            raise ValueError("Cannot delete this view: %s" % view_name)

        if view_name in self:
            self[view_name].delete()
            self.poll()

    def __setitem__(self, view_name, job_names_list):
        new_view = self.create(view_name)
        if isinstance(job_names_list, str):
            job_names_list = [job_names_list]
        for job_name in job_names_list:
            if not new_view.add_job(job_name):
                # Something wrong - delete view
                del self[new_view]
                raise TypeError("Job %s does not exist in Jenkins" % job_name)

    def __getitem__(self, view_name):
        self.poll()
        for row in self._data.get("views", []):
            if row["name"] == view_name:
                return View(row["url"], row["name"], self.jenkins)

        raise KeyError("View %s not found" % view_name)

    def iteritems(self):
        """
        Get the names & objects for all views
        """
        self.poll()
        for row in self._data.get("views", []):
            name = row["name"]
            url = row["url"]

            yield name, View(url, name, self.jenkins)

    def __contains__(self, view_name):
        """
        True if view_name is the name of a defined view
        """
        return view_name in self.keys()

    def iterkeys(self):
        """
        Get the names of all available views
        """
        self.poll()
        for row in self._data.get("views", []):
            yield row["name"]

    def keys(self):
        """
        Return a list of the names of all views
        """
        return list(self.iterkeys())

    def create(self, view_name, view_type=LIST_VIEW, config=None):
        """
        Create a view
        :param view_name: name of new view, str
        :param view_type: type of the view, one of the constants in Views, str
        :param config: XML configuration of the new view
        :return: new View obj or None if view was not created
        """
        log.info('Creating "%s" view "%s"', view_type, view_name)

        if view_name in self:
            log.warning('View "%s" already exists', view_name)
            return self[view_name]

        url = "%s/createView" % self.jenkins.baseurl

        if view_type == self.CATEGORIZED_VIEW:
            if not config:
                raise JenkinsAPIException(
                    "Job XML config cannot be empty for CATEGORIZED_VIEW"
                )

            params = {"name": view_name}

            self.jenkins.requester.post_xml_and_confirm_status(
                url, data=config, params=params
            )
        else:
            headers = {"Content-Type": "application/x-www-form-urlencoded"}
            data = {
                "name": view_name,
                "mode": view_type,
                "Submit": "OK",
                "json": json.dumps({"name": view_name, "mode": view_type}),
            }

            self.jenkins.requester.post_and_confirm_status(
                url, data=data, headers=headers
            )

        self.poll()
        return self[view_name]



"""
Module for jenkinsapi Jenkins object
"""

from __future__ import annotations

import time
import logging
import warnings

from urllib.parse import urlparse
from urllib.request import Request, HTTPRedirectHandler, build_opener
from urllib.parse import quote as urlquote
from urllib.parse import urlencode

from requests import HTTPError, ConnectionError

from jenkinsapi import config
from jenkinsapi.credentials import Credentials
from jenkinsapi.credentials import Credentials2x
from jenkinsapi.credentials import CredentialsById
from jenkinsapi.executors import Executors
from jenkinsapi.jobs import Jobs
from jenkinsapi.job import Job
from jenkinsapi.view import View
from jenkinsapi.label import Label
from jenkinsapi.node import Node
from jenkinsapi.nodes import Nodes
from jenkinsapi.plugins import Plugins
from jenkinsapi.plugin import Plugin
from jenkinsapi.utils.requester import Requester
from jenkinsapi.views import Views
from jenkinsapi.queue import Queue
from jenkinsapi.fingerprint import Fingerprint
from jenkinsapi.jenkinsbase import JenkinsBase
from jenkinsapi.custom_exceptions import JenkinsAPIException
from jenkinsapi.utils.crumb_requester import CrumbRequester


log = logging.getLogger(__name__)


class Jenkins(JenkinsBase):
    """
    Represents a jenkins environment.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        baseurl: str,
        username: str = "",
        password: str = "",
        requester=None,
        lazy: bool = False,
        ssl_verify: bool = True,
        cert=None,
        timeout: int = 10,
        use_crumb: bool = True,
        max_retries=None,
    ) -> None:
        """
        :param baseurl: baseurl for jenkins instance including port, str
        :param username: username for jenkins auth, str
        :param password: password for jenkins auth, str
        :return: a Jenkins obj
        """
        self.username = username
        self.password = password
        if requester is None:
            if use_crumb:
                requester = CrumbRequester
            else:
                requester = Requester

            self.requester = requester(
                username,
                password,
                baseurl=baseurl,
                ssl_verify=ssl_verify,
                cert=cert,
                timeout=timeout,
                max_retries=max_retries,
            )
        else:
            self.requester = requester

        self.requester.timeout = timeout
        self.lazy = lazy
        self.jobs_container = None
        JenkinsBase.__init__(self, baseurl, poll=not lazy)

    def _poll(self, tree=None):
        url = self.python_api_url(self.baseurl)
        return self.get_data(
            url, tree="jobs[name,color,url]" if not tree else tree
        )

    def _poll_if_needed(self):
        if self.lazy and self._data is None:
            self.poll()

    def _clone(self):
        return Jenkins(
            self.baseurl,
            username=self.username,
            password=self.password,
            requester=self.requester,
        )

    def base_server_url(self):
        if config.JENKINS_API in self.baseurl:
            return self.baseurl[: -(len(config.JENKINS_API))]
        return self.baseurl

    def validate_fingerprint(self, id_):
        obj_fingerprint = Fingerprint(self.baseurl, id_, jenkins_obj=self)
        obj_fingerprint.validate()
        log.info(msg="Jenkins says %s is valid" % id_)

    def get_artifact_data(self, id_):
        obj_fingerprint = Fingerprint(self.baseurl, id_, jenkins_obj=self)
        obj_fingerprint.validate()
        return obj_fingerprint.get_info()

    def validate_fingerprint_for_build(self, digest, filename, job, build):
        obj_fingerprint = Fingerprint(self.baseurl, digest, jenkins_obj=self)
        return obj_fingerprint.validate_for_build(filename, job, build)

    def get_jenkins_obj(self):
        return self

    def get_jenkins_obj_from_url(self, url: str):
        return Jenkins(url, self.username, self.password, self.requester)

    def get_create_url(self) -> str:
        # This only ever needs to work on the base object
        return "%s/createItem" % self.baseurl

    def get_nodes_url(self) -> str:
        # This only ever needs to work on the base object
        return self.nodes.baseurl

    @property
    def jobs(self):
        if self.jobs_container is None:
            self.jobs_container = Jobs(self)

        return self.jobs_container

    def get_jobs(self):
        """
        Fetch all the build-names on this Jenkins server.
        """
        return self.jobs.iteritems()

    def get_jobs_info(self):
        """
        Get the jobs information
        :return url, name
        """
        for name, job in self.jobs.iteritems():
            yield job.url, name

    def get_job(self, jobname: str) -> Job:
        """
        Get a job by name
        :param jobname: name of the job, str
        :return: Job obj
        """
        return self.jobs[jobname]

    def get_job_by_url(self, url: str, job_name: str) -> Job:
        """
        Get a job by url
        :param url: jobs' url
        :param jobname: name of the job, str
        :return: Job obj
        """
        return Job(url, job_name, self)

    def has_job(self, jobname: str) -> bool:
        """
        Does a job by the name specified exist
        :param jobname: string
        :return: boolean
        """
        return jobname in self.jobs

    def create_job(self, jobname: str, xml) -> Job:
        """
        Create a job

        alternatively you can create job using Jobs object:
        self.jobs['job_name'] = config
        :param jobname: name of new job, str
        :param config: configuration of new job, xml
        :return: new Job obj
        """
        return self.jobs.create(jobname, xml)

    def create_multibranch_pipeline_job(
        self, jobname, xml, block=True, delay=60
    ):
        """
        :return: list of new Job objects
        """
        return self.jobs.create_multibranch_pipeline(
            jobname, xml, block, delay
        )

    def copy_job(self, jobname: str, newjobname: str) -> Job:
        return self.jobs.copy(jobname, newjobname)

    def build_job(self, jobname: str, params=None) -> None:
        """
        Invoke a build by job name
        :param jobname: name of exist job, str
        :param params: the job params, dict
        :return: none
        """
        self[jobname].invoke(build_params=params or {})

    def delete_job(self, jobname: str) -> None:
        """
        Delete a job by name
        :param jobname: name of a exist job, str
        :return: new jenkins_obj
        """
        del self.jobs[jobname]

    def rename_job(self, jobname: str, newjobname: str) -> Job:
        """
        Rename a job
        :param jobname: name of a exist job, str
        :param newjobname: name of new job, str
        :return: new Job obj
        """
        return self.jobs.rename(jobname, newjobname)

    def items(self):
        """
        :param return: A list of pairs.
            Each pair will be (job name, Job object)
        """
        return list(self.iteritems())

    def get_jobs_list(self):
        return self.jobs.keys()

    def iterkeys(self):
        return self.jobs.iterkeys()

    def iteritems(self):
        return self.jobs.iteritems()

    def keys(self):
        return self.jobs.keys()

    def __str__(self) -> str:
        return "Jenkins server at %s" % self.baseurl

    @property
    def views(self):
        return Views(self)

    def get_view_by_url(self, view_url: str):
        # for nested view
        view_name = view_url.split("/view/")[-1].replace("/", "")
        return View(view_url, view_name, jenkins_obj=self)

    def delete_view_by_url(self, viewurl: str):
        url = f"{viewurl}/doDelete"
        self.requester.post_and_confirm_status(url, data="")
        self.poll()
        return self

    def get_label(self, label_name: str) -> Label:
        label_url = "%s/label/%s" % (self.baseurl, label_name)
        return Label(label_url, label_name, jenkins_obj=self)

    def __getitem__(self, jobname: str) -> Job:
        """
        Get a job by name
        :param jobname: name of job, str
        :return: Job obj
        """
        return self.jobs[jobname]

    def __len__(self) -> int:
        return len(self.jobs)

    def __contains__(self, jobname: str) -> bool:
        """
        Does a job by the name specified exist
        :param jobname: string
        :return: boolean
        """
        return jobname in self.jobs

    def __delitem__(self, job_name: str) -> None:
        del self.jobs[job_name]

    def get_node(self, nodename: str) -> Node:
        """Get a node object for a specific node"""
        return self.nodes[nodename]

    def get_node_url(self, nodename: str = "") -> str:
        """Return the url for nodes"""
        url = urlparse.urljoin(
            self.base_server_url(), "computer/%s" % urlquote(nodename)
        )
        return url

    def get_queue_url(self):
        url = f"{self.base_server_url()}/queue"
        return url

    def get_queue(self) -> Queue:
        queue_url = self.get_queue_url()
        return Queue(queue_url, self)

    def get_nodes(self) -> Nodes:
        return Nodes(self.baseurl, self)

    @property
    def nodes(self):
        return self.get_nodes()

    def has_node(self, nodename: str) -> bool:
        """
        Does a node by the name specified exist
        :param nodename: string, hostname
        :return: boolean
        """
        self.poll()
        return nodename in self.nodes

    def delete_node(self, nodename: str) -> None:
        """
        Remove a node from the managed slave list
        Please note that you cannot remove the master node

        :param nodename: string holding a hostname
        :return: None
        """
        del self.nodes[nodename]

    def create_node(
        self,
        name: str,
        num_executors: int = 2,
        node_description: str = "",
        remote_fs: str = "/var/lib/jenkins",
        labels=None,
        exclusive: bool = False,
    ) -> Node:
        """
        Create a new JNLP slave node by name.

        To create SSH node, please see description in Node class

        :param name: fqdn of slave, str
        :param num_executors: number of executors, int
        :param node_description: a freetext field describing the node
        :param remote_fs: jenkins path, str
        :param labels: labels to associate with slave, str
        :param exclusive: tied to specific job, boolean
        :return: node obj
        """
        node_dict = {
            "num_executors": num_executors,
            "node_description": node_description,
            "remote_fs": remote_fs,
            "labels": labels,
            "exclusive": exclusive,
        }
        return self.nodes.create_node(name, node_dict)

    def create_node_with_config(self, name: str, config) -> Node | None:
        """
        Create a new slave node with specific configuration.
        Config should be resemble the output of node.get_node_attributes()
        :param str name: name of slave
        :param dict config: Node attributes for Jenkins API request
            to create node
            (See function output Node.get_node_attributes())
        :return: node obj
        """
        return self.nodes.create_node_with_config(name=name, config=config)

    def get_plugins_url(self, depth):
        # This only ever needs to work on the base object
        return f"{self.baseurl}/pluginManager/api/python?depth={depth}"

    def install_plugin(
        self,
        plugin: str | Plugin,
        restart: bool = True,
        force_restart: bool = False,
        wait_for_reboot: bool = True,
        no_reboot_warning: bool = False,
    ):
        """
        Install a plugin and optionally restart jenkins.
        @param plugin: Plugin (string or Plugin object) to be installed
        @param restart: Boolean, restart Jenkins when required by plugin
        @param force_restart: Boolean, force Jenkins to restart,
            ignoring plugin preferences
        @param no_warning: Don't show warning when restart is needed and
        restart parameters are set to False
        """
        if not isinstance(plugin, Plugin):
            plugin = Plugin(plugin)
        self.plugins[plugin.shortName] = plugin
        if force_restart or (restart and self.plugins.restart_required):
            self.safe_restart(wait_for_reboot=wait_for_reboot)
        elif self.plugins.restart_required and not no_reboot_warning:
            warnings.warn(
                "System reboot is required, but automatic reboot is disabled. "
                "Please reboot manually."
            )

    def install_plugins(
        self,
        plugin_list,
        restart: bool = True,
        force_restart: bool = False,
        wait_for_reboot: bool = True,
        no_reboot_warning: bool = False,
    ) -> None:
        """
        Install a list of plugins and optionally restart jenkins.
        @param plugin_list: List of plugins (strings, Plugin objects or
            a mix of the two) to be installed
        @param restart: Boolean, restart Jenkins when required by plugin
        @param force_restart: Boolean, force Jenkins to restart,
            ignoring plugin preferences
        """
        plugins = [
            p if isinstance(p, Plugin) else Plugin(p) for p in plugin_list
        ]
        for plugin in plugins:
            self.install_plugin(plugin, restart=False, no_reboot_warning=True)
        if force_restart or (restart and self.plugins.restart_required):
            self.safe_restart(wait_for_reboot=wait_for_reboot)
        elif self.plugins.restart_required and not no_reboot_warning:
            warnings.warn(
                "System reboot is required, but automatic reboot is disabled. "
                "Please reboot manually."
            )

    def delete_plugin(
        self,
        plugin: str | Plugin,
        restart: bool = True,
        force_restart: bool = False,
        wait_for_reboot: bool = True,
        no_reboot_warning: bool = False,
    ) -> None:
        """
        Delete a plugin and optionally restart jenkins. Will not delete
        dependencies.
        @param plugin: Plugin (string or Plugin object) to be deleted
        @param restart: Boolean, restart Jenkins when required by plugin
        @param force_restart: Boolean, force Jenkins to restart,
            ignoring plugin preferences
        """
        if isinstance(plugin, Plugin):
            plugin = plugin.shortName
        del self.plugins[plugin]
        if force_restart or (restart and self.plugins.restart_required):
            self.safe_restart(wait_for_reboot=wait_for_reboot)
        elif self.plugins.restart_required and not no_reboot_warning:
            warnings.warn(
                "System reboot is required, but automatic reboot is disabled. "
                "Please reboot manually."
            )

    def delete_plugins(
        self,
        plugin_list,
        restart: bool = True,
        force_restart: bool = False,
        wait_for_reboot: bool = True,
        no_reboot_warning: bool = False,
    ):
        """
        Delete a list of plugins and optionally restart jenkins. Will not
        delete dependencies.
        @param plugin_list: List of plugins (strings, Plugin objects or
            a mix of the two) to be deleted
        @param restart: Boolean, restart Jenkins when required by plugin
        @param force_restart: Boolean, force Jenkins to restart,
            ignoring plugin preferences
        """
        for plugin in plugin_list:
            self.delete_plugin(plugin, restart=False, no_reboot_warning=True)
        if force_restart or (restart and self.plugins.restart_required):
            self.safe_restart(wait_for_reboot=wait_for_reboot)
        elif self.plugins.restart_required and not no_reboot_warning:
            warnings.warn(
                "System reboot is required, but automatic reboot is disabled. "
                "Please reboot manually."
            )

    def safe_restart(self, wait_for_reboot: bool = True):
        """restarts jenkins when no jobs are running"""
        # NB: unlike other methods, the value of resp.status_code
        # here can be 503 even when everything is normal
        url = "%s/safeRestart" % (self.baseurl,)
        valid = self.requester.VALID_STATUS_CODES + [503, 500]
        resp = self.requester.post_and_confirm_status(
            url, data="", valid=valid
        )
        if wait_for_reboot:
            self._wait_for_reboot()
        return resp

    def _wait_for_reboot(self) -> None:
        # We need to make sure all jobs have finished,
        # and that jenkins is actually restarting.
        # One way to be sure is to make sure jenkins is really down.
        wait = 5
        count = 0
        max_count = 30
        self.__jenkins_is_unavailable()  # Blocks until jenkins is restarting
        while count < max_count:
            time.sleep(wait)
            try:
                self.poll()
                len(self.plugins)  # Make sure jenkins is fully started
                return  # By this time jenkins is back online
            except (HTTPError, ConnectionError):
                msg = (
                    "Jenkins has not restarted yet!  (This is"
                    " try {0} of {1}, waited {2} seconds so far)"
                    "  Sleeping and trying again.."
                )
                msg = msg.format(count, max_count, count * wait)
                log.debug(msg)
            count += 1
        msg = (
            "Jenkins did not come back from safe restart! "
            "Waited %s seconds altogether.  This "
            "failure may cause other failures."
        )
        log.critical(msg, count * wait)

    def __jenkins_is_unavailable(self):
        while True:
            try:
                res = self.requester.get_and_confirm_status(
                    self.baseurl, valid=[503, 500, 200]
                )
                # If there is a running job in Jenkins, the system message will
                # pop up but the Jenkins instance will return 200
                if (
                    res.status_code == 200
                    and "Jenkins is going to shut down"
                    in str(res.content, encoding="utf-8")
                ):
                    time.sleep(1)
                    continue
                return True
            except ConnectionError:
                # This is also a possibility while Jenkins is restarting
                return True
            except HTTPError:
                # This is a return code that is not 503,
                # so Jenkins is likely available
                time.sleep(1)

    def safe_exit(self, wait_for_exit: bool = True, max_wait: int = 360):
        """
        Restarts jenkins when no jobs are running, except for pipeline jobs
        """
        # NB: unlike other methods, the value of resp.status_code
        # here can be 503 even when everything is normal
        url = f"{self.baseurl}/safeExit"
        valid = self.requester.VALID_STATUS_CODES + [503, 500]
        resp = self.requester.post_and_confirm_status(
            url, data="", valid=valid
        )
        if wait_for_exit:
            self._wait_for_exit(max_wait=max_wait)
        return resp

    def _wait_for_exit(self, max_wait: int = 360) -> None:
        # We need to make sure all non pipeline jobs have finished,
        # and that jenkins is unavailable
        self.__jenkins_is_unresponsive(max_wait=max_wait)

    def __jenkins_is_unresponsive(self, max_wait: int = 360):
        # Blocks until jenkins returns ConnectionError or JenkinsAPIException
        # Default wait is one hour
        is_alive = True
        wait = 0
        while is_alive and wait < max_wait:
            try:
                self.requester.get_and_confirm_status(
                    self.baseurl, valid=[200]
                )
                time.sleep(1)
                wait += 1
                is_alive = True
            except (ConnectionError, JenkinsAPIException):
                # Jenkins is finally down
                is_alive = False
                return True
            except HTTPError:
                # This is a return code that is not 503,
                # so Jenkins is likely available, and we need to wait
                time.sleep(1)
                wait += 1
                is_alive = True

    def quiet_down(self):
        """
        Put Jenkins in a Quiet mode, preparation for restart.
        No new builds started
        """
        # NB: unlike other methods, the value of resp.status_code
        # here can be 503 even when everything is normal
        url = "%s/quietDown" % (self.baseurl,)
        valid = self.requester.VALID_STATUS_CODES + [503, 500]
        resp = self.requester.post_and_confirm_status(
            url, data="", valid=valid
        )
        return resp

    def cancel_quiet_down(self):
        """Cancel the effect of the quiet-down command"""
        # NB: unlike other methods, the value of resp.status_code
        # here can be 503 even when everything is normal
        url = "%s/cancelQuietDown" % (self.baseurl,)
        valid = self.requester.VALID_STATUS_CODES + [503, 500]
        resp = self.requester.post_and_confirm_status(
            url, data="", valid=valid
        )
        return resp

    @property
    def plugins(self):
        return self.get_plugins()

    def get_plugins(self, depth: int = 1) -> Plugins:
        url = self.get_plugins_url(depth=depth)
        return Plugins(url, self)

    def has_plugin(self, plugin_name: str) -> bool:
        return plugin_name in self.plugins

    def get_executors(self, nodename: str) -> Executors:
        url = f"{self.baseurl}/computer/{nodename}"
        return Executors(url, nodename, self)

    def get_master_data(self):
        url = f"{self.baseurl}/computer/api/python"
        return self.get_data(url)

    @property
    def version(self) -> str:
        """
        Return version number of Jenkins
        """
        response = self.requester.get_and_confirm_status(self.baseurl)
        version_key = "X-Jenkins"
        return response.headers.get(version_key, "0.0")

    def get_credentials(self, cred_class=Credentials2x):
        """
        Return credentials
        """

        if "credentials" not in self.plugins:
            raise JenkinsAPIException("Credentials plugin not installed")

        if self.plugins["credentials"].version.startswith("1."):
            url = f"{self.baseurl}/credential-store/domain/_/"
            return Credentials(url, self)

        url = f"{self.baseurl}/credentials/store/system/domain/_/"
        return cred_class(url, self)

    @property
    def credentials(self):
        return self.get_credentials(Credentials2x)

    @property
    def credentials_by_id(self):
        return self.get_credentials(CredentialsById)

    @property
    def is_quieting_down(self) -> bool:
        url = "%s/api/python?tree=quietingDown" % (self.baseurl,)
        data = self.get_data(url=url)
        return data.get("quietingDown", False)

    def shutdown(self) -> None:
        url = "%s/exit" % self.baseurl
        self.requester.post_and_confirm_status(url, data="")

    def generate_new_api_token(
        self, new_token_name: str = "Token By jenkinsapi python"
    ):
        subUrl = (
            "/me/descriptorByName/jenkins.security."
            "ApiTokenProperty/generateNewToken"
        )
        url = "%s%s" % (self.baseurl, subUrl)
        data = urlencode({"newTokenName": new_token_name})
        response = self.requester.post_and_confirm_status(url, data=data)
        token = response.json()["data"]["tokenValue"]
        return token

    def run_groovy_script(self, script: str) -> str:
        """
        Runs the requested groovy script on the Jenkins server returning the
        result as text.
        Raises a JenkinsAPIException if the returned HTTP response code from
        the POST request is not 200 OK.

        Example:

            server = Jenkins(...)
            script = 'println "Hello world!"'
            result = server.run_groovy_script(script)
            print(result) # will print "Hello world!"
        """
        url = f"{self.baseurl}/scriptText"
        data = urlencode({"script": script})

        response = self.requester.post_and_confirm_status(url, data=data)
        if response.status_code != 200:
            raise JenkinsAPIException(
                "Unexpected response %d." % response.status_code
            )

        return response.text

    def use_auth_cookie(self) -> None:
        assert self.username and self.baseurl, (
            "Please provide jenkins url, username "
            "and password to get the session ID cookie."
        )

        login_url = "j_acegi_security_check"
        jenkins_url = "{0}/{1}".format(self.baseurl, login_url)
        data = urlencode(
            {"j_username": self.username, "j_password": self.password}
        ).encode("utf-8")

        class SmartRedirectHandler(HTTPRedirectHandler):
            def extract_cookie(self, setcookie):
                # Extracts the last cookie.
                # Example of set-cookie value for python2
                # ('set-cookie', 'JSESSIONID.30blah=blahblahblah;Path=/;
                #   HttpOnly, JSESSIONID.30ablah=blahblah;Path=/;HttpOnly'),
                return setcookie.split(",")[-1].split(";")[0].strip("\n\r ")

            def http_error_302(self, req, fp, code, msg, headers):
                # Jenkins can send several Set-Cookie values sometimes
                #  The valid one is the last one
                for header, value in headers.items():
                    if header.lower() == "set-cookie":
                        cookie = self.extract_cookie(value)

                req.headers["Cookie"] = cookie
                result = HTTPRedirectHandler.http_error_302(
                    self, req, fp, code, msg, headers
                )
                result.orig_status = code
                result.orig_headers = headers
                result.cookie = cookie
                return result

        request = Request(jenkins_url, data)
        opener = build_opener(SmartRedirectHandler())
        res = opener.open(request)
        Requester.AUTH_COOKIE = res.cookie



import os
import time
import shutil
import logging
import datetime
import tempfile
import posixpath
import requests
import queue
import threading
import tarfile
import subprocess
from urllib3 import Retry
from urllib.parse import urlparse

from requests.adapters import HTTPAdapter

from jenkinsapi.jenkins import Jenkins
from jenkinsapi.custom_exceptions import JenkinsAPIException

log = logging.getLogger(__name__)


class FailedToStart(Exception):
    pass


class TimeOut(Exception):
    pass


class StreamThread(threading.Thread):
    def __init__(self, name, q, stream, fn_log):
        threading.Thread.__init__(self)
        self.name = name
        self.queue = q
        self.stream = stream
        self.fn_log = fn_log
        self._stop = threading.Event()

    def stop(self):
        self._stop.set()

    def stopped(self):
        return self._stop.isSet()

    def run(self):
        log.info("Starting %s", self.name)

        while True:
            if self._stop.is_set():
                break
            line = self.stream.readline()
            if line:
                self.fn_log(line.rstrip())
                self.queue.put((self.name, line))
            else:
                break

        self.queue.put((self.name, None))


class JenkinsLancher(object):
    """
    Launch jenkins
    """

    JENKINS_WEEKLY_WAR_URL = "http://get.jenkins.io/war/latest/jenkins.war"
    JENKINS_LTS_WAR_URL = (
        "https://get.jenkins.io/war-stable/latest/jenkins.war"
    )

    def __init__(
        self,
        local_orig_dir,
        systests_dir,
        war_name,
        plugin_urls=None,
        jenkins_url=None,
    ):
        if jenkins_url is not None:
            self.jenkins_url = jenkins_url
            self.http_port = urlparse(jenkins_url).port
            self.start_new_instance = False
        else:
            import socket

            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.bind(("", 0))
            sock.listen(1)
            port = sock.getsockname()[1]
            sock.close()
            self.http_port = port
            self.jenkins_url = "http://localhost:%s" % self.http_port
            self.start_new_instance = True

        self.threads = []
        self.war_path = os.path.join(local_orig_dir, war_name)
        self.local_orig_dir = local_orig_dir
        self.systests_dir = systests_dir
        self.war_filename = war_name

        if "JENKINS_HOME" not in os.environ:
            self.jenkins_home = tempfile.mkdtemp(prefix="jenkins-home-")
            os.environ["JENKINS_HOME"] = self.jenkins_home
        else:
            self.jenkins_home = os.environ["JENKINS_HOME"]

        self.jenkins_process = None
        self.queue = queue.Queue()
        self.plugin_urls = plugin_urls or []
        if os.environ.get("JENKINS_VERSION", "stable") == "stable":
            self.JENKINS_WAR_URL = self.JENKINS_LTS_WAR_URL
        else:
            self.JENKINS_WAR_URL = self.JENKINS_WEEKLY_WAR_URL

    def update_war(self):
        os.chdir(self.systests_dir)
        if os.path.exists(self.war_path):
            log.info(
                "War file already present, delete it to redownload and"
                " update jenkins"
            )
        else:
            log.info("Downloading Jenkins War")
            script_dir = os.path.join(self.systests_dir, "get-jenkins-war.sh")
            subprocess.check_call(
                [
                    script_dir,
                    self.JENKINS_WAR_URL,
                    self.local_orig_dir,
                    self.war_filename,
                ]
            )

    def update_config(self):
        from jenkinsapi_tests import systests

        file = os.path.join(
            os.path.dirname(systests.__file__), "jenkins_home.tar.gz"
        )

        with open(file, "rb") as f:
            with tarfile.open(fileobj=f, mode="r:gz") as tarball:
                tarball.extractall(path=self.jenkins_home)

    def install_plugins(self):
        plugin_dest_dir = os.path.join(self.jenkins_home, "plugins")
        log.info("Plugins will be installed in '%s'", plugin_dest_dir)

        if not os.path.exists(plugin_dest_dir):
            os.mkdir(plugin_dest_dir)

        for url in self.plugin_urls:
            self.install_plugin(url, plugin_dest_dir)

    def install_plugin(self, hpi_url, plugin_dest_dir):
        sess = requests.Session()
        adapter = HTTPAdapter(
            max_retries=Retry(total=5, backoff_factor=1, allowed_methods=None)
        )
        sess.mount("http://", adapter)
        sess.mount("https://", adapter)

        path = urlparse(hpi_url).path
        filename = posixpath.basename(path)
        plugin_orig_dir = os.path.join(self.local_orig_dir, "plugins")
        if not os.path.exists(plugin_orig_dir):
            os.mkdir(plugin_orig_dir)
        plugin_orig_path = os.path.join(plugin_orig_dir, filename)
        plugin_dest_path = os.path.join(plugin_dest_dir, filename)
        if os.path.exists(plugin_orig_path):
            log.info(
                "%s already locally present, delete the file to redownload"
                " and update",
                filename,
            )
        else:
            log.info("Downloading %s from %s", filename, hpi_url)
            with sess.get(hpi_url, stream=True) as hget:
                hget.raise_for_status()
                with open(plugin_orig_path, "wb") as hpi:
                    for chunk in hget.iter_content(chunk_size=8192):
                        hpi.write(chunk)
        log.info("Installing %s", filename)
        shutil.copy(plugin_orig_path, plugin_dest_path)
        # Create an empty .pinned file, so that the downloaded plugin
        # will be used, instead of the version bundled in jenkins.war
        # See https://wiki.jenkins-ci.org/display/JENKINS/Pinned+Plugins
        open(plugin_dest_path + ".pinned", "a").close()

    def stop(self):
        if self.start_new_instance:
            log.info("Shutting down jenkins.")
            # Start the threads
            for thread in self.threads:
                thread.stop()

            Jenkins(self.jenkins_url).shutdown()
            # self.jenkins_process.terminate()
            # self.jenkins_process.wait()
            # Do not remove jenkins home if JENKINS_URL is set
            if "JENKINS_URL" not in os.environ:
                shutil.rmtree(self.jenkins_home, ignore_errors=True)
            log.info("Jenkins stopped.")

    def block_until_jenkins_ready(self, timeout):
        start_time = datetime.datetime.now()
        timeout_time = start_time + datetime.timedelta(seconds=timeout)

        while True:
            try:
                Jenkins(self.jenkins_url)
                log.info("Jenkins is finally ready for use.")
            except JenkinsAPIException:
                log.info("Jenkins is not yet ready...")
            if datetime.datetime.now() > timeout_time:
                raise TimeOut("Took too long for Jenkins to become ready...")
            time.sleep(5)

    def start(self, timeout=60):
        if self.start_new_instance:
            self.jenkins_home = os.environ.get(
                "JENKINS_HOME", self.jenkins_home
            )
            self.update_war()
            self.update_config()
            self.install_plugins()

            os.chdir(self.local_orig_dir)

            jenkins_command = [
                "java",
                "-Djenkins.install.runSetupWizard=false",
                "-Dhudson.DNSMultiCast.disabled=true",
                "-jar",
                self.war_filename,
                "--httpPort=%d" % self.http_port,
            ]

            log.info("About to start Jenkins...")
            log.info("%s> %s", os.getcwd(), " ".join(jenkins_command))
            self.jenkins_process = subprocess.Popen(
                jenkins_command,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            self.threads = [
                StreamThread(
                    "out", self.queue, self.jenkins_process.stdout, log.info
                ),
                StreamThread(
                    "err", self.queue, self.jenkins_process.stderr, log.warning
                ),
            ]

            # Start the threads
            for thread in self.threads:
                thread.start()

            while True:
                try:
                    streamName, line = self.queue.get(
                        block=True, timeout=timeout
                    )
                    # Python 3.x
                    if isinstance(line, bytes):
                        line = line.decode("UTF-8")
                except queue.Empty:
                    log.warning("Input ended unexpectedly")
                    break
                else:
                    if line:
                        if "Failed to initialize Jenkins" in line:
                            raise FailedToStart(line)

                        if "Invalid or corrupt jarfile" in line:
                            raise FailedToStart(line)

                        if "is fully up and running" in line:
                            log.info(line)
                            return
                    else:
                        log.warning("Stream %s has terminated", streamName)

            self.block_until_jenkins_ready(timeout)


if __name__ == "__main__":
    logging.basicConfig()
    logging.getLogger("").setLevel(logging.INFO)

    log.info("Hello!")

    jl = JenkinsLancher(
        "/home/aleksey/src/jenkinsapi_lechat/jenkinsapi_tests"
        "/systests/localinstance_files",
        "/home/aleksey/src/jenkinsapi_lechat/jenkinsapi_tests/systests",
        "jenkins.war",
    )

    jl.start()
    log.info("Jenkins was launched...")

    time.sleep(10)

    log.info("...now to shut it down!")
    jl.stop()
